"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Setting_1 = require("../../models/Setting");
const CommandService_1 = require("../CommandService");
const locale_1 = require("../../locale");
const KvStore_1 = require("../KvStore");
class SpellCheckerService {
    constructor() {
        this.latestSelectedLanguages_ = [];
    }
    static instance() {
        if (this.instance_)
            return this.instance_;
        this.instance_ = new SpellCheckerService();
        return this.instance_;
    }
    initialize(driver) {
        return __awaiter(this, void 0, void 0, function* () {
            this.driver_ = driver;
            this.latestSelectedLanguages_ = yield this.loadLatestSelectedLanguages();
            this.setupDefaultLanguage();
            this.applyStateToDriver();
        });
    }
    get defaultLanguage() {
        return 'en-US';
    }
    loadLatestSelectedLanguages() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield KvStore_1.default.instance().value('spellCheckerService.latestSelectedLanguages');
            if (!result)
                return [];
            return JSON.parse(result);
        });
    }
    addLatestSelectedLanguage(language) {
        return __awaiter(this, void 0, void 0, function* () {
            const languages = this.latestSelectedLanguages_.slice();
            if (languages.length > 5)
                languages.splice(0, 1);
            if (languages.includes(language)) {
                languages.splice(languages.indexOf(language), 1);
            }
            languages.splice(0, 0, language);
            this.latestSelectedLanguages_ = languages;
            yield KvStore_1.default.instance().setValue('spellCheckerService.latestSelectedLanguages', JSON.stringify(this.latestSelectedLanguages_));
        });
    }
    setupDefaultLanguage() {
        if (!Setting_1.default.value('spellChecker.language')) {
            const l = this.driver_.language;
            this.setLanguage(l ? l : this.defaultLanguage);
        }
    }
    get availableLanguages() {
        return this.driver_.availableLanguages;
    }
    applyStateToDriver() {
        this.driver_.setLanguage(this.enabled ? this.language : '');
    }
    setLanguage(language) {
        Setting_1.default.setValue('spellChecker.language', language);
        this.applyStateToDriver();
        void this.addLatestSelectedLanguage(language);
    }
    get language() {
        return Setting_1.default.value('spellChecker.language');
    }
    get enabled() {
        return Setting_1.default.value('spellChecker.enabled');
    }
    toggleEnabled() {
        Setting_1.default.toggle('spellChecker.enabled');
        this.applyStateToDriver();
    }
    addToDictionary(language, word) {
        return __awaiter(this, void 0, void 0, function* () {
            this.driver_.addWordToSpellCheckerDictionary(language, word);
        });
    }
    contextMenuItems(misspelledWord, dictionarySuggestions) {
        if (!misspelledWord)
            return [];
        const output = [];
        output.push({ type: 'separator' });
        if (dictionarySuggestions.length) {
            for (const suggestion of dictionarySuggestions) {
                output.push({
                    label: suggestion,
                    click: () => {
                        void CommandService_1.default.instance().execute('replaceMisspelling', suggestion);
                    },
                });
            }
        }
        else {
            output.push({
                label: `(${(0, locale_1._)('No suggestions')})`,
                enabled: false,
                click: () => { },
            });
        }
        output.push({ type: 'separator' });
        output.push({
            label: (0, locale_1._)('Add to dictionary'),
            click: () => {
                void this.addToDictionary(this.language, misspelledWord);
            },
        });
        return output;
    }
    changeLanguageMenuItem(language, enabled, checked) {
        return {
            label: (0, locale_1.countryDisplayName)(language),
            type: 'radio',
            checked: checked,
            enabled: enabled,
            click: () => {
                this.setLanguage(language);
            },
        };
    }
    changeLanguageMenuItems(selectedLanguage, enabled) {
        const languageMenuItems = [];
        for (const locale of this.driver_.availableLanguages) {
            languageMenuItems.push(this.changeLanguageMenuItem(locale, enabled, locale === selectedLanguage));
        }
        languageMenuItems.sort((a, b) => {
            return a.label < b.label ? -1 : +1;
        });
        return languageMenuItems;
    }
    spellCheckerConfigMenuItems(selectedLanguage, useSpellChecker) {
        const latestLanguageItems = this.latestSelectedLanguages_.map((language) => {
            return this.changeLanguageMenuItem(language, true, language === selectedLanguage);
        });
        if (latestLanguageItems.length)
            latestLanguageItems.splice(0, 0, { type: 'separator' });
        latestLanguageItems.sort((a, b) => {
            return a.label < b.label ? -1 : +1;
        });
        return [
            {
                label: (0, locale_1._)('Use spell checker'),
                type: 'checkbox',
                checked: useSpellChecker,
                click: () => {
                    this.toggleEnabled();
                },
            },
            ...latestLanguageItems,
            {
                type: 'separator',
            },
            // Can be removed once it does work
            // {
            // 	label: '⚠ Spell checker doesn\'t work in Markdown editor ⚠',
            // 	enabled: false,
            // },
            {
                type: 'separator',
            },
            {
                label: (0, locale_1._)('Change language'),
                submenu: this.changeLanguageMenuItems(selectedLanguage, useSpellChecker),
            },
        ];
    }
    spellCheckerConfigMenuItem(selectedLanguage, useSpellChecker) {
        return {
            label: (0, locale_1._)('Spell checker'),
            submenu: this.spellCheckerConfigMenuItems(selectedLanguage, useSpellChecker),
        };
    }
}
exports.default = SpellCheckerService;
//# sourceMappingURL=SpellCheckerService.js.map