"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LockType = void 0;
const types_1 = require("./utils/types");
const shim_1 = require("../../shim");
const JoplinError_1 = require("../../JoplinError");
const time_1 = require("../../time");
const { fileExtension, filename } = require('../../path-utils');
var LockType;
(function (LockType) {
    LockType["None"] = "";
    LockType["Sync"] = "sync";
    LockType["Exclusive"] = "exclusive";
})(LockType = exports.LockType || (exports.LockType = {}));
function defaultAcquireLockOptions() {
    return {
        clearExistingSyncLocksFromTheSameClient: false,
        timeoutMs: 0,
    };
}
class LockHandler {
    constructor(api, options = null) {
        this.api_ = null;
        this.refreshTimers_ = {};
        this.autoRefreshInterval_ = 1000 * 60;
        this.lockTtl_ = 1000 * 60 * 3;
        if (!options)
            options = {};
        this.api_ = api;
        if ('lockTtl' in options)
            this.lockTtl_ = options.lockTtl;
        if ('autoRefreshInterval' in options)
            this.autoRefreshInterval_ = options.autoRefreshInterval;
    }
    get lockTtl() {
        return this.lockTtl_;
    }
    // Should only be done for testing purposes since all clients should
    // use the same lock max age.
    set lockTtl(v) {
        this.lockTtl_ = v;
    }
    lockFilename(lock) {
        return `${[lock.type, lock.clientType, lock.clientId].join('_')}.json`;
    }
    lockTypeFromFilename(name) {
        const ext = fileExtension(name);
        if (ext !== 'json')
            return LockType.None;
        if (name.indexOf(LockType.Sync) === 0)
            return LockType.Sync;
        if (name.indexOf(LockType.Exclusive) === 0)
            return LockType.Exclusive;
        return LockType.None;
    }
    lockFilePath(lock) {
        return `${types_1.Dirnames.Locks}/${this.lockFilename(lock)}`;
    }
    lockFileToObject(file) {
        const p = filename(file.path).split('_');
        return {
            type: p[0],
            clientType: p[1],
            clientId: p[2],
            updatedTime: file.updated_time,
        };
    }
    locks(lockType = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.api_.list(types_1.Dirnames.Locks);
            if (result.hasMore)
                throw new Error('hasMore not handled'); // Shouldn't happen anyway
            const output = [];
            for (const file of result.items) {
                const type = this.lockTypeFromFilename(file.path);
                if (type === LockType.None)
                    continue;
                if (lockType && type !== lockType)
                    continue;
                const lock = this.lockFileToObject(file);
                output.push(lock);
            }
            return output;
        });
    }
    lockIsActive(lock, currentDate) {
        return currentDate.getTime() - lock.updatedTime < this.lockTtl;
    }
    hasActiveLock(lockType, clientType = null, clientId = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const lock = yield this.activeLock(lockType, clientType, clientId);
            return !!lock;
        });
    }
    // Finds if there's an active lock for this clientType and clientId and returns it.
    // If clientType and clientId are not specified, returns the first active lock
    // of that type instead.
    activeLock(lockType, clientType = null, clientId = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const locks = yield this.locks(lockType);
            const currentDate = yield this.api_.remoteDate();
            if (lockType === LockType.Exclusive) {
                const activeLocks = locks
                    .slice()
                    .filter((lock) => this.lockIsActive(lock, currentDate))
                    .sort((a, b) => {
                    if (a.updatedTime === b.updatedTime) {
                        return a.clientId < b.clientId ? -1 : +1;
                    }
                    return a.updatedTime < b.updatedTime ? -1 : +1;
                });
                if (!activeLocks.length)
                    return null;
                const activeLock = activeLocks[0];
                if (clientType && clientType !== activeLock.clientType)
                    return null;
                if (clientId && clientId !== activeLock.clientId)
                    return null;
                return activeLock;
            }
            else if (lockType === LockType.Sync) {
                for (const lock of locks) {
                    if (clientType && lock.clientType !== clientType)
                        continue;
                    if (clientId && lock.clientId !== clientId)
                        continue;
                    if (this.lockIsActive(lock, currentDate))
                        return lock;
                }
                return null;
            }
            throw new Error(`Unsupported lock type: ${lockType}`);
        });
    }
    saveLock(lock) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.api_.put(this.lockFilePath(lock), JSON.stringify(lock));
        });
    }
    // This is for testing only
    saveLock_(lock) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.saveLock(lock);
        });
    }
    acquireSyncLock(clientType, clientId) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let isFirstPass = true;
                while (true) {
                    const [exclusiveLock, syncLock] = yield Promise.all([
                        this.activeLock(LockType.Exclusive),
                        this.activeLock(LockType.Sync, clientType, clientId),
                    ]);
                    if (exclusiveLock) {
                        throw new JoplinError_1.default(`Cannot acquire sync lock because the following client has an exclusive lock on the sync target: ${this.lockToClientString(exclusiveLock)}`, 'hasExclusiveLock');
                    }
                    if (syncLock) {
                        // Normally the second pass should happen immediately afterwards, but if for some reason
                        // (slow network, etc.) it took more than 10 seconds then refresh the lock.
                        if (isFirstPass || Date.now() - syncLock.updatedTime > 1000 * 10) {
                            yield this.saveLock(syncLock);
                        }
                        return syncLock;
                    }
                    // Something wrong happened, which means we saved a lock but we didn't read
                    // it back. Could be application error or server issue.
                    if (!isFirstPass)
                        throw new Error('Cannot acquire sync lock: either the lock could be written but not read back. Or it was expired before it was read again.');
                    yield this.saveLock({
                        type: LockType.Sync,
                        clientType: clientType,
                        clientId: clientId,
                    });
                    isFirstPass = false;
                }
            }
            catch (error) {
                yield this.releaseLock(LockType.Sync, clientType, clientId);
                throw error;
            }
        });
    }
    lockToClientString(lock) {
        return `(${lock.clientType} #${lock.clientId})`;
    }
    acquireExclusiveLock(clientType, clientId, options = null) {
        return __awaiter(this, void 0, void 0, function* () {
            // The logic to acquire an exclusive lock, while avoiding race conditions is as follow:
            //
            // - Check if there is a lock file present
            //
            // - If there is a lock file, see if I'm the one owning it by checking that its content has my identifier.
            // - If that's the case, just write to the data file then delete the lock file.
            // - If that's not the case, just wait a second or a small random length of time and try the whole cycle again-.
            //
            // -If there is no lock file, create one with my identifier and try the whole cycle again to avoid race condition (re-check that the lock file is really mine)-.
            options = Object.assign(Object.assign({}, defaultAcquireLockOptions()), options);
            const startTime = Date.now();
            function waitForTimeout() {
                return __awaiter(this, void 0, void 0, function* () {
                    if (!options.timeoutMs)
                        return false;
                    const elapsed = Date.now() - startTime;
                    if (options.timeoutMs && elapsed < options.timeoutMs) {
                        yield time_1.default.sleep(2);
                        return true;
                    }
                    return false;
                });
            }
            try {
                while (true) {
                    const [activeSyncLock, activeExclusiveLock] = yield Promise.all([
                        this.activeLock(LockType.Sync),
                        this.activeLock(LockType.Exclusive),
                    ]);
                    if (activeSyncLock) {
                        if (options.clearExistingSyncLocksFromTheSameClient && activeSyncLock.clientId === clientId && activeSyncLock.clientType === clientType) {
                            yield this.releaseLock(LockType.Sync, clientType, clientId);
                        }
                        else {
                            if (yield waitForTimeout())
                                continue;
                            throw new JoplinError_1.default(`Cannot acquire exclusive lock because the following clients have a sync lock on the target: ${this.lockToClientString(activeSyncLock)}`, 'hasSyncLock');
                        }
                    }
                    if (activeExclusiveLock) {
                        if (activeExclusiveLock.clientId === clientId) {
                            // Save it again to refresh the timestamp
                            yield this.saveLock(activeExclusiveLock);
                            return activeExclusiveLock;
                        }
                        else {
                            // If there's already an exclusive lock, wait for it to be released
                            if (yield waitForTimeout())
                                continue;
                            throw new JoplinError_1.default(`Cannot acquire exclusive lock because the following client has an exclusive lock on the sync target: ${this.lockToClientString(activeExclusiveLock)}`, 'hasExclusiveLock');
                        }
                    }
                    else {
                        // If there's not already an exclusive lock, acquire one
                        // then loop again to check that we really got the lock
                        // (to prevent race conditions)
                        yield this.saveLock({
                            type: LockType.Exclusive,
                            clientType: clientType,
                            clientId: clientId,
                        });
                        yield time_1.default.msleep(100);
                    }
                }
            }
            catch (error) {
                yield this.releaseLock(LockType.Exclusive, clientType, clientId);
                throw error;
            }
        });
    }
    autoLockRefreshHandle(lock) {
        return [lock.type, lock.clientType, lock.clientId].join('_');
    }
    startAutoLockRefresh(lock, errorHandler) {
        const handle = this.autoLockRefreshHandle(lock);
        if (this.refreshTimers_[handle]) {
            throw new Error(`There is already a timer refreshing this lock: ${handle}`);
        }
        this.refreshTimers_[handle] = {
            id: null,
            inProgress: false,
        };
        this.refreshTimers_[handle].id = shim_1.default.setInterval(() => __awaiter(this, void 0, void 0, function* () {
            if (this.refreshTimers_[handle].inProgress)
                return;
            const defer = () => {
                if (!this.refreshTimers_[handle])
                    return;
                this.refreshTimers_[handle].inProgress = false;
            };
            this.refreshTimers_[handle].inProgress = true;
            let error = null;
            const hasActiveLock = yield this.hasActiveLock(lock.type, lock.clientType, lock.clientId);
            if (!this.refreshTimers_[handle])
                return defer(); // Timeout has been cleared
            if (!hasActiveLock) {
                // If the previous lock has expired, we shouldn't try to acquire a new one. This is because other clients might have performed
                // in the meantime operations that invalidates the current operation. For example, another client might have upgraded the
                // sync target in the meantime, so any active operation should be cancelled here. Or if the current client was upgraded
                // the sync target, another client might have synced since then, making any cached data invalid.
                // In some cases it should be safe to re-acquire a lock but adding support for this would make the algorithm more complex
                // without much benefits.
                error = new JoplinError_1.default('Lock has expired', 'lockExpired');
            }
            else {
                try {
                    yield this.acquireLock(lock.type, lock.clientType, lock.clientId);
                    if (!this.refreshTimers_[handle])
                        return defer(); // Timeout has been cleared
                }
                catch (e) {
                    error = e;
                }
            }
            if (error) {
                if (this.refreshTimers_[handle]) {
                    shim_1.default.clearInterval(this.refreshTimers_[handle].id);
                    delete this.refreshTimers_[handle];
                }
                errorHandler(error);
            }
            defer();
        }), this.autoRefreshInterval_);
        return handle;
    }
    stopAutoLockRefresh(lock) {
        const handle = this.autoLockRefreshHandle(lock);
        if (!this.refreshTimers_[handle]) {
            // Should not throw an error because lock may have been cleared in startAutoLockRefresh
            // if there was an error.
            // throw new Error(`There is no such lock being auto-refreshed: ${this.lockToString(lock)}`);
            return;
        }
        shim_1.default.clearInterval(this.refreshTimers_[handle].id);
        delete this.refreshTimers_[handle];
    }
    acquireLock(lockType, clientType, clientId, options = null) {
        return __awaiter(this, void 0, void 0, function* () {
            options = Object.assign(Object.assign({}, defaultAcquireLockOptions()), options);
            if (lockType === LockType.Sync) {
                return this.acquireSyncLock(clientType, clientId);
            }
            else if (lockType === LockType.Exclusive) {
                return this.acquireExclusiveLock(clientType, clientId, options);
            }
            else {
                throw new Error(`Invalid lock type: ${lockType}`);
            }
        });
    }
    releaseLock(lockType, clientType, clientId) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.api_.delete(this.lockFilePath({
                type: lockType,
                clientType: clientType,
                clientId: clientId,
            }));
        });
    }
}
exports.default = LockHandler;
//# sourceMappingURL=LockHandler.js.map